#include "kct.h"

/*	#define DEBUG	*/
/*	#define DEBUG_DIGEST	*/
/*	#define DEBUG_AFS_TKT	*/
/*	#define DEBUG_TIMESTAMP	*/


/*	#define SINGLE_PROCESS	*/

#ifdef UOF_ID_CHECK
#define UOF_MAX_ID_LEN 8
#define UOF_MIN_ID_LEN 3
#endif

#define QUEUE_LEN 20
#define MAX_ID_LEN 1024
#define MAX_REALM_LEN 1024
#define MAX_BUF_SIZE 1024
#define MAX_MEMORY_ALLOC 1000000

int  TGS_PORT = 5039;
int  SINGLE_PROCESS = 0;
char TGS_KEYTAB[MAX_KEYTAB_NAME_LEN] = "/opt/tgs+/tgs+.keytab";
char TGS_PRINC_NAME[MAX_BUF_SIZE] = "tgs+";
char TGS_VERSION[MAX_BUF_SIZE] = "tgs version 1";
char CA_FILE[MAX_KEYTAB_NAME_LEN] = "/opt/tgs+/cert/ca-bundle.crt";
char CA_PATH[MAX_KEYTAB_NAME_LEN] = "/opt/tgs+/cert";

#define VALID_TIME 10 /** this refers to 10mins that ssl handshake is valid **/
#define TKT_LIFETIME 300 /** currently set to 5 minutes **/

#define TIMESLICE 300

#define NO_ERROR            0
#define INTERNAL_ERROR      244
#define ERR_RD_PRIV         245
#define INVALID_HANDSHAKE   242
#define OLD_HANDSHAKE       243
#define INVALID_ID          244
#define REALM_MISMATCH      245
#define COMMUNICATION_ERROR 246

#define s2l(c,s)        ((s=(unsigned long)(((c[0])&0xff))<<24| \
                            (unsigned long)(((c[1])&0xff)<<16)|\
                            (unsigned long)(((c[2])&0xff)<<8)| \
                            (unsigned long)(((c[3])   )  &0xff)), c+=4)

typedef struct saved_hndshake {
  int version;

  unsigned char *master_key;
  unsigned int master_key_length;

  unsigned char *c_h;
  unsigned int c_h_l;

  unsigned char *s_h;
  unsigned int s_h_l;

  unsigned char *s_c;
  unsigned int s_c_l;

  unsigned char *s_c_r;
  unsigned int s_c_r_l;

  unsigned char *s_k_e;
  unsigned int s_k_e_l;

  unsigned char *c_c;
  unsigned int c_c_l;

  unsigned char *c_k_e;
  unsigned int c_k_e_l;

  unsigned char *c_v;
  unsigned int c_v_l;

} SAVED_HANDSHAKE;

void                     *handle;

static int handle_request(tgs_context *);
static int handle_ap_req(tgs_context *);
static int receive_handshake(tgs_context *, SAVED_HANDSHAKE *);
static int verify_handshake(tgs_context *, SAVED_HANDSHAKE *);
static int generate_service_ticket(tgs_context *, unsigned char **, int *);
static int send_reply(int, long, unsigned char *, int);

static void cert_verify_mac(unsigned char *, int, EVP_MD_CTX *, 
			    unsigned char *);
static void calculate_digest(SAVED_HANDSHAKE *, unsigned char *);

static int retrieve_cert(int, unsigned char *, unsigned int, EVP_PKEY **, 
			 int *, X509 **);
static int verify_cert(X509 *);
static unsigned char *read_message(unsigned char **, unsigned int *);
static int verify_client_signature(unsigned char *, unsigned int, 
				   unsigned char *, EVP_PKEY *);
static int verify_client_timestamp(unsigned char *);
static int retrieve_identity(tgs_context *, char **, char **, X509 *);
static int check_server_identity(tgs_context *);
static void process_exit(int, const struct sigaction *, struct sigaction *);
static int get_service_key(krb5_context, krb5_keyblock *, krb5_key_data **,
			   krb5_principal);
static int cb_local(int ok, X509_STORE_CTX *ctx);
static void usage();
static void read_config_file(char *);

void usage() {
  fprintf(stderr, "tgs [list of parameters] or -f [configuration file]\n");
  fprintf(stderr, "List of parameters:\n");
  fprintf(stderr, "-s (enables single process mode of operation)\n");
  fprintf(stderr, "-p [port]\n");
  fprintf(stderr, "-k [keytab file name]\n");
  fprintf(stderr, "-pname [tgs principal name]\n");
  fprintf(stderr, "-CAfile [CA bundle file name]\n");
  fprintf(stderr, "-CApath [path to dir with certs]\n");

  exit(1);
}

void read_config_file(char *filename) {
  FILE *f = fopen(filename, "r");
  char op[MAX_BUF_SIZE], value[MAX_BUF_SIZE];

  if(f == NULL) {
    fprintf(stderr, "couldn't open configuration file %s\n", filename);
    exit(1);
  }

  while(fscanf(f, "%s %s", op, value) != EOF) {
    if(!strcmp(op, "TGS_PORT")) {
      TGS_PORT = atoi(value);
    }
    else if(!strcmp(op, "TGS_KEYTAB")) {
      strcpy(TGS_KEYTAB, value);
    }
    else if(!strcmp(op, "TGS_PRINC_NAME")) {
      strcpy(TGS_PRINC_NAME, value);
    }
    else if(!strcmp(op, "CA_FILE")) {
      strcpy(CA_FILE, value);
    }
    else if(!strcmp(op, "CA_PATH")) {
      strcpy(CA_PATH, value);
    }
    else if(!strcmp(op, "SINGLE_PROCESS")) {
      SINGLE_PROCESS = 1;
    }
    else {
      fprintf(stderr, "unknwon option %s\n", op);
      goto end;
    }
  }

 end:
  fclose(f);
}

/**
 ** TGS+: 
 ** 0. reads in configuration parameters either from the console or 
 **    a configuration file
 ** 1. creates a listening socket for incoming request
 ** 2. upon receiving a message, it forks of a new process to handle the
 **    request. request is handled by the function 'handle_request'
 **/

int main(int argc, char *argv[]) {

  int                  sock, csock;
  struct sockaddr_in   srvaddr, clntaddr;
  struct hostent       *phost;
  int                  size = 1;
  char                 *hname;
  int                  c_port;
  tgs_context          *request;
  int badop = 0;
  char                 CONF_FILE_NAME[MAX_KEYTAB_NAME_LEN];
  krb5_error_code      retval;

  argc--;
  argv++;
  while (argc >= 1) {
    if(!strcmp(*argv, "-port")) {
      if(--argc < 1) goto bad;
      TGS_PORT = atoi(*(++argv));
    }
    else if(!strcmp(*argv, "-k")) {
      if(--argc < 1) goto bad;
      strcpy(TGS_KEYTAB, *(++argv));
    }
    else if(!strcmp(*argv, "-pname")) {
      if(--argc < 1) goto bad;
      strcpy(TGS_PRINC_NAME, *(++argv));
    }
    else if(!strcmp(*argv, "-CAfile")) {
      if(--argc < 1) goto bad;
      strcpy(CA_FILE, *(++argv));
    }
    else if(!strcmp(*argv, "-CApath")) {
      if(--argc < 1) goto bad;
      strcpy(CA_PATH, *(++argv));
    }
    else if(!strcmp(*argv, "-f")) {
      if(--argc < 1) goto bad;
      strcpy(CONF_FILE_NAME, *(++argv));
      read_config_file(CONF_FILE_NAME);
    }
    else if(!strcmp(*argv, "-s")) {
      SINGLE_PROCESS = 1;
    }
    else {
      badop = 1;
      fprintf(stderr, "unknown option %s\n", *argv);
    }

    argc--;
    argv++;
  }

  if(badop) {
  bad:
    usage();
  }

  open_log();

  if(SINGLE_PROCESS) {
    if ((retval = kadm5_init(TGS_PRINC_NAME, NULL, KADM5_ADMIN_SERVICE, NULL,
			KADM5_STRUCT_VERSION, KADM5_API_VERSION_2,
			&handle))) {
      com_err("tgs", retval, "initializing kadm5 library");
      exit(1);
    }
  }

#if defined(sun)
  sighold(SIGPIPE);
#else
  sigblock(SIGPIPE);
#endif
  bzero(&srvaddr, sizeof(srvaddr));
  bzero(&clntaddr, sizeof(clntaddr));

  if((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
    perror("opening stream socket");
    return -1;
  }

  if(setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&size,sizeof(size)) < 0) {
    perror("setsockopt failed");
    return -1;
  }

  srvaddr.sin_family = AF_INET;
  srvaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  srvaddr.sin_port = htons(TGS_PORT);

  if(bind(sock, (struct sockaddr *) &srvaddr, sizeof(srvaddr)) == -1) {
    perror("binding stream socket");
    close(sock);
    return -1;
  }

  listen(sock, QUEUE_LEN);

  size = sizeof(clntaddr);

  while(1) {

    request = (tgs_context *)get_memory(1, sizeof(tgs_context));

    if ((request->ws_sock = accept(sock, (struct sockaddr *) &clntaddr, 
				   (int *) &size)) == -1) {
      perror("accept failed");
      continue;
    }

#ifdef INFO
    if((phost = gethostbyaddr((char *)&clntaddr.sin_addr, 
			      sizeof(struct in_addr), 
			      clntaddr.sin_family)) == NULL) {
      hname = (char *)get_memory(strlen(inet_ntoa(clntaddr.sin_addr))+1, 
				 sizeof(char));
    } else {
      hname = (char *)get_memory(strlen(phost->h_name)+1, sizeof(char));
    }

    strcpy(hname, inet_ntoa(clntaddr.sin_addr));    
    c_port = ntohs(clntaddr.sin_port);

    log(TGSLOG_DEBUG, "connection accepted from %s port %d (@130)\n", 
	request->ws_hname, request->ws_port);

    free(hname);
#endif

    if(SINGLE_PROCESS) {
      handle_request(request);
    }
    else {
      switch( fork() ) {
      case -1: /** fork failed **/
        fprintf(stderr, "fork failed %s\n", strerror(errno));
        close(request->ws_sock);
        close(sock);
        free_memory(request);
        exit(-1);
      case 0: /** child process **/
        handle_request(request);
        exit(1);
      default: break; /** parent process **/
      }
    }
  }
  close(sock);

  close_log();
  
cleanup:
  if(SINGLE_PROCESS) 
    kadm5_destroy(handle);

}

/**
 ** This the main function that handles TGS+ requests
 ** before processing the request, a process sets up a timer that will 
 ** interrupt this process in TIMESLICE amount of time and kill the process.
 ** I've added this so that I wouldn't have dangling processing
 **
 ** 1. TGS+ handles kerberos AP_REQ. If AP_REQ fails, it returns from this
 ** function and closes the connection to the client.
 ** 
 ** 2. Receives SSL handshake bytes. SSL_session key is send encrypted so it
 ** calls krb5_rd_priv. It's sent encrypted with the session key. If 
 ** decryption of this message fails, TGS+ returns an error message back to
 ** the client. 
 **
 ** 3. After a successful read of the handshake, TGS+ calls the procedure to 
 ** verify the handshake. Verification consists of the following
 ** - matches identity in server's certificate against the identity used
 **   during the AP_REQ
 ** - verifies signature of both client's and server's certificate. it checks
 **   that the certificate has been signed by the CA TGS+ trusts
 ** - verifies client's signature inside of SSL handhshake (this code is
 **   exactly the same as the verification code in openssl)
 ** - checks that the client's timestamp in client hello message is not too
 **   old
 ** If at any point verification fails, TGS+ sends back an error message back
 ** to the client.
 ** 
 ** 4. After the handshake been successfully verified, TGS+ calls the routine
 ** to create a service ticket, currently AFS ticket. At this point if
 ** something goes wrong, then it's an internal error and client is notified. 
 ** A ticket is created for the identity inside of the client's certificate.
 **
 ** 5. If TGS+ was successful in creating a service ticket, it sends it 
 ** back to the client with NO_ERROR error message.
 **/

static int handle_request(tgs_context *context) {

  krb5_error_code retval = 0;
  SAVED_HANDSHAKE *sh;
  struct itimerval itimer;
  struct sigaction act;
  unsigned char *reply=NULL;
  int len = 0;

  /** Set up a timer. It will terminate this process after 5mins
   ** This is done so that we don't have hanging processes.
   **/
  sigemptyset(&act.sa_mask);
  act.sa_handler = ( void (*) () ) process_exit;
  act.sa_flags = SA_RESTART;
  sigaction(SIGVTALRM, &act, NULL);
 
  itimer.it_value.tv_sec = TIMESLICE;
  itimer.it_value.tv_usec = 0; 
  itimer.it_interval.tv_sec = TIMESLICE; 
  itimer.it_interval.tv_usec = 0; 

  if(setitimer(ITIMER_VIRTUAL, &itimer, NULL) != 0) {
    perror("setting interval timer");
    log(TGSLOG_CRIT, "unable to set the timer\n");
    return INTERNAL_ERROR;
  }

  /** Second, initialize krb structures and receive/validate AP_REQ
   ** Do a check on requester's identity
   **/
  retval = handle_ap_req(context);
  switch(retval) {
  case NO_ERROR: break;
  default: goto send_reply;
  }

  /** Allocate memory for the ssl handshake that we are about to receive
   ** from the web server
   **/
  sh = (SAVED_HANDSHAKE *)get_memory(1, sizeof(SAVED_HANDSHAKE));

  /** Receive and parse the handshake **/
  retval = receive_handshake(context, sh);
  switch(retval) {
  case NO_ERROR: break;    
  default: goto send_reply;
  }

  /** Verify the handshake **/
  retval = verify_handshake(context, sh);
  switch(retval) {
  case NO_ERROR: break;
  default: goto send_reply;
  }

  /** Create and send AFS ticket **/
  /** Ticket is going to be issued to the identity inside of the client's
   ** certificate present in SSL handshake.
   **
   ** Currently, tickets are issued only for AFS service. In the future,
   ** the requester should specify which service it needs a ticket for
   **/
  retval = generate_service_ticket(context, &reply, &len);
  switch(retval) {
  case NO_ERROR: break;
  default: goto send_reply;
  }

 send_reply:
  send_reply(context->ws_sock, retval, reply, len);

 cleanup:

  if(reply && len > 0)          free_memory(reply); 
  if(sh != NULL) {
    if(sh->c_h)        free_memory(sh->c_h);
    if(sh->s_h)        free_memory(sh->s_h);
    if(sh->s_c)        free_memory(sh->s_c);
    if(sh->s_c_r)      free_memory(sh->s_c_r);
    if(sh->s_k_e && sh->s_k_e_l > 0)      free_memory(sh->s_k_e);
    if(sh->c_c)        free_memory(sh->c_c);
    if(sh->c_k_e)      free_memory(sh->c_k_e);
    if(sh->c_v)        free_memory(sh->c_v);
    if(sh->master_key) free_memory(sh->master_key);
    free_memory(sh);
  }
  
/*
  if(context->ws_ticket && context->ws_ticket->enc_part2)
    krb5_free_enc_tkt_part(context->krb_context, 
			   context->ws_ticket->enc_part2);
 */
  if(context->ws_ticket)
    krb5_free_ticket(context->krb_context, context->ws_ticket);
  if(context->ws_auth_context)
    krb5_auth_con_free(context->krb_context, context->ws_auth_context);
  if(context->tgs_server)
    krb5_free_principal(context->krb_context, context->tgs_server);
  if(context->krb_context) 
    krb5_free_context(context->krb_context);
  
  if(context->client_id)    free_memory(context->client_id);
  if(context->client_realm) free_memory(context->client_realm);
  if(context->ws_id)        free_memory(context->ws_id);
  if(context->ws_realm)     free_memory(context->ws_realm);
  close(context->ws_sock);
  
  free_memory(context);

  return retval;
}

static int handle_ap_req(tgs_context *context) {

  char *client_name;
  krb5_error_code retval;
  krb5_enctype enctype;
  krb5_keytab_entry ktent;

#ifdef DEBUG
  fprintf(stderr, "handle_ap_req: start\n");
#endif

  if((retval = krb5_init_context(&context->krb_context)) != 0) {
    com_err("tgs_server", retval, "unable to initialize krb5");
    goto cleanup;
  }
  
  if((retval = krb5_auth_con_init(context->krb_context, 
				  &context->ws_auth_context)) != 0) {
    com_err("tgs_server", retval, "unable to initialize auth_context\n");
    goto cleanup;
  }

  if((retval = krb5_kt_resolve(context->krb_context, TGS_KEYTAB, 
			       &context->tgs_keytab)) != 0) {
    com_err("tgs_server", retval, "unable to resolve keytable file %s", 
	    TGS_KEYTAB);
    goto cleanup;
  }

  if((retval = krb5_parse_name(context->krb_context, TGS_PRINC_NAME,
			       &context->tgs_server)) != 0) {
    com_err("tgs_server", retval, "unable to generate service principal %s", 
	    TGS_PRINC_NAME);
    goto cleanup;
  }

  if((retval = krb5_recvauth(context->krb_context, &context->ws_auth_context,
			     (krb5_pointer)&context->ws_sock, TGS_VERSION, 
			     context->tgs_server, 0, context->tgs_keytab, 
			     &context->ws_ticket)) != 0) {
     com_err("tgs_server", retval, "recvauth failed");
     goto cleanup;
  }

  enctype = context->ws_ticket->enc_part.enctype;
  if((retval = krb5_kt_get_entry(context->krb_context, context->tgs_keytab, 
				 context->ws_ticket->server,
				 context->ws_ticket->enc_part.kvno, enctype,
				 &ktent)) != 0) {
    com_err("tgs_server", retval, "unable to retrieve a keytab entry");
    goto cleanup;
  }

  if((retval = krb5_decrypt_tkt_part(context->krb_context, &ktent.key, 
				     context->ws_ticket)) != 0) {
    com_err("tgs_server", retval, "unable to decrypt the ticket");
    goto cleanup;
  }

  if((retval = krb5_unparse_name(context->krb_context, 
				 context->ws_ticket->enc_part2->client, 
				 &client_name)) != 0) {
    com_err("tgs_server", retval, "unable to parse the name in the tkt");
    goto cleanup;
  }

  context->ws_id = (char *)get_memory(MAX_ID_LEN, sizeof(char));
  context->ws_realm = (char *)get_memory(MAX_REALM_LEN, sizeof(char));
  sscanf(client_name, "%[^@]@%s", context->ws_id, context->ws_realm);
  
  free_memory(client_name); 
  
#ifdef DEBUG
  fprintf(stderr, "1: client id=%s client realm=%s\n", context->ws_id, 
	  context->ws_realm);
#endif

#ifdef TGS_WS_REALM_MATCH
  /** I'm not sure this check is needed. I'm checking to make sure that web
   ** server is in the same Kerberos realm as the TGS+. However, it might 
   ** not be true if cross-realm authentication ever take off
   **/
  {
    char *default_realm;
    if((retval = krb5_get_default_realm(context->krb_context, 
					&default_realm)) != 0) {
      com_err("SystemX", retval, "couldn't get the default realm");
      return retval;
    }

    if(strcasecmp(context->ws_realm, default_realm)) {
      log(TGSLOG_CRIT, "1. received ticket for invalid realm %s\n", 
	  context->ws_realm);
      return REALM_MISMATCH;
    }

    free_memory(default_realm);
  }
#endif  

#ifdef DEBUG
  fprintf(stderr, "handle_ap_req: end\n");
#endif

 cleanup:
  if(context->tgs_keytab)
    krb5_kt_close(context->krb_context, context->tgs_keytab);
  
  return retval;
}

static int receive_handshake(tgs_context *context, SAVED_HANDSHAKE *sh) {

  int sock = context->ws_sock;

  unsigned long xmitlen;
  unsigned char *recv_buf;
  unsigned char *p;
  int len = sizeof(unsigned long);

  krb5_data recv_data, recv_unenc_data;
  krb5_error_code retval = 0;

  krb5_address remote_kaddr;
  struct sockaddr_in remote_addr;
  int addrlen;

#ifdef DEBUG
  fprintf(stderr, "receive_handshake: start\n");
#endif

  if(recvblock(sock, (unsigned char *)&xmitlen, sizeof(xmitlen)) < 0) {
    perror("socket closed");
    close(sock);
    return COMMUNICATION_ERROR;
  }
  xmitlen = ntohl(xmitlen);

  /** check that header is within reasonable range **/
  if(xmitlen <= 0 || xmitlen > MAX_MEMORY_ALLOC) {
    fprintf(stderr, "recieve_handshake: requested more memory than allowed\n");
    return INVALID_HANDSHAKE;
  }

#ifdef DEBUG
  fprintf(stderr, "whole receive buffer: %d\n", xmitlen);
#endif

  p = recv_buf = (unsigned char *)
    get_memory(xmitlen, sizeof(unsigned char));
  if(recvblock(sock, (unsigned char *)recv_buf, xmitlen) < 0) {
    perror("socket closed");
    close(sock);
    return COMMUNICATION_ERROR;
  }

  memcpy((char *)&sh->version, p, len);
  sh->version = ntohl(sh->version);
  p += len;

  /** check that version is valid **/
  if(sh->version != TLS1_VERSION && sh->version != SSL3_VERSION){
    fprintf(stderr, "receive_handshake: invalid SSL version\n");
    return INVALID_HANDSHAKE;
  }

  /** all but ServerKeyExchange messages must be present in the handshake
   ** presence of ServerKeyExchange is optional
   **/
  sh->c_h   = read_message(&p, &sh->c_h_l);
  if(sh->c_h_l <= 0 || sh->c_h == NULL) {
    fprintf(stderr, "receive_handshake: invalid CLIENT_HELLO\n");
    return INVALID_HANDSHAKE;
  }
  sh->s_h   = read_message(&p, &sh->s_h_l);
  if(sh->s_h_l <= 0 || sh->s_h == NULL) {
    fprintf(stderr, "receive_handshake: invalid SERVER_HELLO\n");
    return INVALID_HANDSHAKE;
  }
  sh->s_c   = read_message(&p, &sh->s_c_l);
  if(sh->s_c_l <= 0 || sh->s_c == NULL) {
    fprintf(stderr, "receive_handshake: invalid SERVER_CERT\n");
    return INVALID_HANDSHAKE;
  }
  sh->s_c_r = read_message(&p, &sh->s_c_r_l);
  if(sh->s_c_r_l <= 0 || sh->s_c_r == NULL) {
    fprintf(stderr, "receive_handshake: invalid SERVER_CERT_REQ\n");
    return INVALID_HANDSHAKE;
  }
  sh->s_k_e = read_message(&p, &sh->s_k_e_l);
  sh->c_c   = read_message(&p, &sh->c_c_l);
  if(sh->c_c_l <= 0 || sh->c_c == NULL) {
    fprintf(stderr, "receive_handshake: invalid CLIENT_CERT\n");
    return INVALID_HANDSHAKE;
  }
  sh->c_k_e = read_message(&p, &sh->c_k_e_l);
  if(sh->c_k_e_l <= 0 || sh->c_k_e == NULL) {
    fprintf(stderr, "receive_handshake: invalid CLIENT_KEY_EX\n");
    return INVALID_HANDSHAKE;
  }
  sh->c_v   = read_message(&p, &sh->c_v_l);
  if(sh->c_v_l <= 0 || sh->c_v == NULL) {
    fprintf(stderr, "receive_handshake: invalid CLIENT_VERIFY\n");
    return INVALID_HANDSHAKE;
  }

  if(recvblock(sock, (unsigned char *)&recv_data.length, 
	       sizeof(xmitlen)) < 0) {
    perror("socket closed");
    close(sock);
    return COMMUNICATION_ERROR;
  }
  recv_data.length = ntohl(recv_data.length);

  /** check that header is within reasonable range **/
  if(recv_data.length <= 0 || recv_data.length > MAX_MEMORY_ALLOC){
    fprintf(stderr, "receive_handshake: requested more memory than allowed\n");
    return INVALID_HANDSHAKE;
  }

  recv_data.data = (char *)
    get_memory(recv_data.length, sizeof(unsigned char));
  if(recvblock(sock, (unsigned char *)recv_data.data, recv_data.length) < 0) {
    perror("socket closed");
    close(sock);
    return COMMUNICATION_ERROR;
  }
  
  if (retval = krb5_auth_con_setflags(context->krb_context, 
				      context->ws_auth_context,
				      KRB5_AUTH_CONTEXT_DO_SEQUENCE)) {
    com_err("tgs_server", retval, "unable to set authcontext flags");
    retval = INTERNAL_ERROR;
    goto cleanup;
  }

  memset((char *)&remote_addr, 0, sizeof(remote_addr));
  addrlen = sizeof(remote_addr);
  if (getpeername(context->ws_sock, (struct sockaddr *)&remote_addr, 
		  &addrlen) < 0) {
    fprintf(stderr, "getpeername failed\n");
    retval = INTERNAL_ERROR;
    goto cleanup;
  }

  remote_kaddr.addrtype = ADDRTYPE_INET;
  remote_kaddr.length = sizeof(remote_addr.sin_addr);
  remote_kaddr.contents = (krb5_octet *) &remote_addr.sin_addr;
  
  if ((retval = krb5_auth_con_setaddrs(context->krb_context, 
				       context->ws_auth_context, NULL,
				       &remote_kaddr)) != 0) {
    com_err("tgs_server", retval, "unable to set krb addr");
    retval = INTERNAL_ERROR;
    goto cleanup;
  }
  
  remote_kaddr.addrtype = ADDRTYPE_IPPORT;
  remote_kaddr.length = sizeof(remote_addr.sin_port);
  remote_kaddr.contents = (krb5_octet *)&remote_addr.sin_port;

  if((retval = krb5_auth_con_setports(context->krb_context, 
				      context->ws_auth_context, NULL,
				      &remote_kaddr)) != 0) {
    com_err("tgs_server", retval, "unable to set krb port");
    retval = INTERNAL_ERROR;
    goto cleanup;
  }

  if(retval = krb5_rd_priv(context->krb_context, context->ws_auth_context,
			   &recv_data, &recv_unenc_data, NULL)) {
    com_err("tgs_server", retval, "while verifying PRIV message");
    retval = ERR_RD_PRIV;
    goto cleanup;
  }

  p = (unsigned char *)recv_unenc_data.data;
  s2l(p, xmitlen);
  sh->master_key_length = ntohl(xmitlen);
  sh->master_key = (unsigned char *)get_memory(sh->master_key_length, 
					       sizeof(unsigned char));
  memcpy(sh->master_key, p, sh->master_key_length);
  p += sh->master_key_length;
  s2l(p, xmitlen);
  len = ntohl(xmitlen);
  context->service_id = (char *)get_memory(len, sizeof(char));
  memcpy(context->service_id, p,len);
  
#ifdef DEBUG
  print_buffer(sh->master_key, sh->master_key_length, "master_key: ");
#endif
     
 cleanup:
   if(recv_buf)        free_memory(recv_buf); 

#ifdef DEBUG
   fprintf(stderr, "receive_handshake: end\n");
#endif
  return retval;
}

static int verify_handshake(tgs_context *context, SAVED_HANDSHAKE *s) {

  unsigned char *verify_md=NULL;
  X509 *client_cert=NULL, *server_cert=NULL;
  EVP_PKEY *client_pkey=NULL, *server_pkey=NULL;
  int client_type=0, server_type=0, retval = 0;

  verify_md = (unsigned char *)
    get_memory(EVP_MAX_MD_SIZE*5, sizeof(unsigned char));

  calculate_digest(s, verify_md);

  server_pkey = (EVP_PKEY *)get_memory(1, sizeof(EVP_PKEY));
  retval = retrieve_cert(s->version, s->s_c + 4, s->s_c_l - 4,
			 &server_pkey, &server_type, &server_cert);
  switch(retval) {
  case NO_ERROR: break;
  default: goto cleanup;
  }

  client_pkey = (EVP_PKEY *)get_memory(1, sizeof(EVP_PKEY));
  retval = retrieve_cert(s->version, s->c_c + 4, s->c_c_l - 4,
			      &client_pkey, &client_type, &client_cert);
  switch(retval) {
  case NO_ERROR: break;
  default: goto cleanup;
  }

  retval = verify_client_signature(s->c_v+4,s->c_v_l-4,verify_md,client_pkey);
  switch(retval) {
  case NO_ERROR: break;
  default: goto cleanup;
  }

  retval = verify_client_timestamp(s->c_h+6);
  switch(retval) {
  case NO_ERROR: break;
  default: goto cleanup;
  }

  retval = retrieve_identity(context, &context->client_id, 
			     &context->client_realm, client_cert);
  switch(retval) {
  case NO_ERROR: break;
  default: goto cleanup;
  }

#ifdef UOFM_ID_CHECK
  {
    int idlen;
    idlen = strlen(context->client->id);
    /** check if id is with in the specified range **/
    if(idlen < UOFM_MIN_ID_LEN || idlen > UOFM_MAX_ID_LEN) {
      retval = INVALID_ID;
      goto cleanup;
    }
  }
#endif

  retval = retrieve_identity(context, &context->ws_id, 
			     &context->ws_realm, server_cert);
  switch(retval) {
  case NO_ERROR: break;
  default: goto cleanup;
  }

  retval = check_server_identity(context);

  switch(retval) {
  case NO_ERROR: break;
  default: goto cleanup;
  }

 cleanup:
  
  if(verify_md)   free_memory(verify_md);
  if(server_cert) X509_free(server_cert);
  if(client_cert) X509_free(client_cert); 
  if(client_pkey) EVP_PKEY_free(client_pkey);
  if(server_pkey) EVP_PKEY_free(server_pkey); 
  
  return retval;

}

/** this function checks that the identity inside of the server certificate
 ** used during the handshake, is the same as the identity inside of the
 ** ticket used in AP_REQ
 **/
static int check_server_identity(tgs_context *context) {

  if(!strncmp(context->ws_id, 
	      context->ws_ticket->enc_part2->client->data->data,  
	      context->ws_ticket->enc_part2->client->data->length))
    return NO_ERROR;
  else {
    fprintf(stderr, "check_server_identity: check failed\n");
    return INVALID_HANDSHAKE;
  }
}

/** this function retrieves a key from the db for the specified service
 **/
static int get_service_key(krb5_context context, 
			   krb5_keyblock *encrypting_key, 
			   krb5_key_data **server_key,
			   krb5_principal service) {

  void                     *handle;

  krb5_error_code          retval = 0;
  kadm5_principal_ent_rec  server;

  if(!SINGLE_PROCESS) {
    if ((retval = kadm5_init(TGS_PRINC_NAME, NULL, KADM5_ADMIN_SERVICE, NULL,
			   KADM5_STRUCT_VERSION, KADM5_API_VERSION_2,
			   &handle))) {
      com_err("tgs", retval, "initializing kadm5 library");
    goto cleanup;
    }
  }


  if ((retval = kadm5_get_principal(handle, service, &server,
				    KADM5_KEY_DATA))) {
    com_err("tgs", retval, "while getting a principal");
    if(!SINGLE_PROCESS)
      goto cleanup;
    else
      return retval;
  }
 
  if ((retval = kadm5_decrypt_key(handle,
				  &server,
				  -1 /*ENCTYPE_DES_CBC_CRC*/,
				  -1,
				  0,
				  encrypting_key, NULL, NULL))) {
    kadm5_free_principal_ent(handle, &server);
    com_err("tgs", retval, "while decrypting the key"); 
    if(!SINGLE_PROCESS)
      goto cleanup;
    else
      return retval;
  }
  
  (*server_key) = server.key_data;

cleanup:
  if (!SINGLE_PROCESS)
    kadm5_destroy(handle);

  return retval;
}

static int generate_service_ticket(tgs_context *tgs_context, 
				   unsigned char **reply_buf, int *len) {

  krb5_context           context = tgs_context->krb_context;
  krb5_principal         service;
  krb5_principal         client;
  krb5_keyblock          encrypting_key;
  krb5_key_data          *server_key;

  krb5_timestamp         tgs_time;
  krb5_keyblock          session_key;

  krb5_data              *response;
  krb5_kdc_rep           reply;
  krb5_enc_kdc_rep_part  reply_encpart;
  krb5_ticket            ticket_reply;
  krb5_enc_tkt_part      enc_tkt_reply;
  krb5_error_code        retval = 0;

  krb5_enctype           useenctype = ENCTYPE_DES_CBC_CRC;
  krb5_last_req_entry    *nolrarray[2], nolrentry;
  int errcode;
  const char	*status = 0;
  krb5_data empty_str = {0, 0, ""};
  
  session_key.contents = 0;

  if((retval = krb5_parse_name(context, tgs_context->client_id, 
			       &client)) != 0) {
    com_err("tgs_server", retval, "unable to create %s service principal", 
	    tgs_context->client_id);
    retval = INTERNAL_ERROR;
    goto cleanup;
  }

  if((retval = krb5_parse_name(context, tgs_context->service_id, 
			       &service)) != 0) {
    com_err("tgs_server", retval, "unable to create %s service principal", 
	    tgs_context->service_id);
    retval = INTERNAL_ERROR;
    goto cleanup;
  }

  retval = get_service_key(context, &encrypting_key, &server_key, service);
  switch(retval) {
  case NO_ERROR: break;
  default: return retval;
  }

  if ((retval = krb5_timeofday(context, &tgs_time))) {
    com_err("tgs_server", retval, "time of day failed\n");
    retval = INTERNAL_ERROR;
    goto cleanup;
  }
    
  if(retval = krb5_c_make_random_key(context, useenctype, &session_key)) {
    com_err("tgs_server", retval, "creating a random key failed\n");
    retval = INTERNAL_ERROR;
    goto cleanup;
  }

  ticket_reply.server = service; /* XXX careful for realm... */
  
#ifdef DEBUG_AFS_TKT
  fprintf(stderr, "client %d id %s %d realm %s\n", client->data->length, 
	  client->data->data, client->realm.length, client->realm.data);
  fprintf(stderr, "server name [%d] %s realm [%d] %s\n", 
	  ticket_reply.server->data->length, ticket_reply.server->data->data, 
	  ticket_reply.server->realm.length, ticket_reply.server->realm.data);
  print_buffer(session_key.contents, session_key.length, "afs key: ");
#endif

  enc_tkt_reply.flags = 0;
  enc_tkt_reply.caddrs = tgs_context->ws_ticket->enc_part2->caddrs;
  reply_encpart.caddrs = 0;
  enc_tkt_reply.times.authtime = tgs_time;
  enc_tkt_reply.times.starttime = tgs_time;
  enc_tkt_reply.times.endtime = tgs_time + TKT_LIFETIME; 
  enc_tkt_reply.times.renew_till = 0;    
  enc_tkt_reply.authorization_data = 0;

  enc_tkt_reply.session = &session_key;
  enc_tkt_reply.client = client;
  enc_tkt_reply.transited.tr_type = KRB5_DOMAIN_X500_COMPRESS;
  enc_tkt_reply.transited.tr_contents = empty_str; /* equivalent of "" */

  ticket_reply.enc_part2 = &enc_tkt_reply;

  if((retval = krb5_encrypt_tkt_part(context, &encrypting_key,
				     &ticket_reply)) != 0) {
    com_err("tgs_server", retval, "encrypting of tkt part failed\n");
    retval = INTERNAL_ERROR;
    goto cleanup;
  }

  ticket_reply.enc_part.kvno = server_key->key_data_kvno;

  /* Start assembling the response */
  reply.msg_type = KRB5_TGS_REP;
  reply.padata = 0;		/* always */
  reply.client = client;
  reply.enc_part.kvno = 0;		/* We are using the session key */
  reply.ticket = &ticket_reply;

  reply_encpart.msg_type = KRB5_TGS_REP;
  reply_encpart.session = &session_key;
  reply_encpart.nonce = 0;
  reply_encpart.times = enc_tkt_reply.times;
  reply_encpart.times.authtime = enc_tkt_reply.times.authtime;
  
  nolrentry.lr_type = KRB5_LRQ_NONE;
  nolrentry.value = 0;
  nolrarray[0] = &nolrentry;
  nolrarray[1] = 0;
  reply_encpart.last_req = nolrarray;	/* not available for TGS reqs */
  reply_encpart.key_exp = 0;		/* ditto */
  reply_encpart.flags = enc_tkt_reply.flags;
  reply_encpart.server = ticket_reply.server;
  reply.enc_part.enctype = tgs_context->ws_ticket->enc_part.enctype;

#ifdef DEBUG_AFS_TKT
  print_buffer(tgs_context->ws_ticket->enc_part2->session->contents,
	       tgs_context->ws_ticket->enc_part2->session->length,
	       "tgs session key: ");
#endif

  if(retval = krb5_encode_kdc_rep(context, KRB5_TGS_REP, &reply_encpart, 0,
       tgs_context->ws_ticket->enc_part2->session, &reply, &response)) {
    com_err("tgs_server", retval, "encoding kdc rep failed\n");
    retval = INTERNAL_ERROR;
    goto cleanup;
  }

  (*len) = response->length;
  (*reply_buf) = (unsigned char*)
    get_memory(response->length, sizeof(unsigned char));
  memcpy((*reply_buf), response->data, response->length);
  
  memset(ticket_reply.enc_part.ciphertext.data, 0,
	 ticket_reply.enc_part.ciphertext.length);
  free_memory(ticket_reply.enc_part.ciphertext.data);
  memset(reply.enc_part.ciphertext.data, 0,
	 reply.enc_part.ciphertext.length);
  free_memory(reply.enc_part.ciphertext.data);
  

  fprintf(stderr, "[%d] TGS+_REQ: ISSUE: authtime %d, %s for %s\n", 
	  getpid(), tgs_time, tgs_context->client_id, tgs_context->service_id);
 cleanup:

  if(service)        krb5_free_principal(context, service);
  if(client)         krb5_free_principal(context, client);
  if(server_key)     krb5_free_key_data_contents(context, server_key);
  if(response)       krb5_free_data(context, response);

  return retval;
  
}

/** sending the reply to the client
 ** first sending a header, telling the client how long the reply is
 ** then attaching the error number in the reply (putting it before the
 ** the service ticket, if there is one)
 **/
static int send_reply(int sock, long retval, unsigned char *reply, int len) {

  long xmitlen = 0;
  unsigned char *send_buf = NULL;

  send_buf = (unsigned char *)
    get_memory(len+sizeof(long), sizeof(unsigned char));
  xmitlen = htonl(retval);
  memcpy(send_buf, (char *)&xmitlen, sizeof(long));
  if(len) memcpy(send_buf+sizeof(long), reply, len);

  xmitlen = htonl(len+sizeof(long));

#ifdef DEBUG
  print_buffer(send_buf, len+sizeof(long), "send buffer:\n");
#endif
  
  if(sendblock(sock, (unsigned char *)&xmitlen, sizeof(xmitlen)) < 0 ) {
    perror("socket closed");
    close(sock);
    return COMMUNICATION_ERROR;
  }
  if(sendblock(sock, (unsigned char *)send_buf, len+sizeof(long)) < 0 ) {
    perror("socket closed");
    close(sock);
    return COMMUNICATION_ERROR;
  }
}

/** to verify client's signature inside of the certificate, we need to
 ** calculate the digest of all handshake messages and verify that the 
 ** digest is the same as the signed digest with in client verify message
 **/
static void calculate_digest(SAVED_HANDSHAKE *s, unsigned char *p) {

  EVP_MD_CTX d1;
  EVP_MD_CTX d2;
  EVP_MD_CTX ctx1;
  const EVP_MD *md5;
  const EVP_MD *sha1;
  unsigned int ret;

#ifdef DEBUG
  fprintf(stderr, "3.1: calculating the digest\n");
#endif

  EVP_add_digest(EVP_md5());
  EVP_add_digest(EVP_sha1());
  EVP_add_digest_alias(SN_md5, "md5");
  EVP_add_digest_alias(SN_sha1, "sha1");
  md5 = EVP_get_digestbyname("md5");
  sha1 = EVP_get_digestbyname("sha1");

  EVP_DigestInit(&d1, md5);
  EVP_DigestInit(&d2, sha1);
  
  if(s->c_h != NULL) {
    EVP_DigestUpdate(&d1, s->c_h, s->c_h_l);
    EVP_DigestUpdate(&d2, s->c_h, s->c_h_l);
  }
  if(s->s_h != NULL) {
    EVP_DigestUpdate(&d1, s->s_h, s->s_h_l);
    EVP_DigestUpdate(&d2, s->s_h, s->s_h_l);
  }
  if(s->s_c != NULL) {
    EVP_DigestUpdate(&d1, s->s_c, s->s_c_l);
    EVP_DigestUpdate(&d2, s->s_c, s->s_c_l);
  }
  if(s->s_k_e != NULL && s->s_k_e_l > 0) {
    EVP_DigestUpdate(&d1, s->s_k_e, s->s_k_e_l);
    EVP_DigestUpdate(&d2, s->s_k_e, s->s_k_e_l);
  }
  if(s->s_c_r != NULL) {
    EVP_DigestUpdate(&d1, s->s_c_r, s->s_c_r_l);
    EVP_DigestUpdate(&d2, s->s_c_r, s->s_c_r_l);
  }
  if(s->c_c != NULL) {
    EVP_DigestUpdate(&d1, s->c_c, s->c_c_l);
    EVP_DigestUpdate(&d2, s->c_c, s->c_c_l);
  }
  if(s->c_k_e != NULL) {
    EVP_DigestUpdate(&d1, s->c_k_e, s->c_k_e_l);
    EVP_DigestUpdate(&d2, s->c_k_e, s->c_k_e_l);
  }

  if(s->version == TLS1_VERSION) {
    EVP_MD_CTX_copy(&ctx1, &d1);
    EVP_DigestFinal(&ctx1, &p[0], &ret);
    
    EVP_MD_CTX_copy(&ctx1, &d2);
    EVP_DigestFinal(&ctx1, &p[MD5_DIGEST_LENGTH], &ret);
  }
  else {
    cert_verify_mac(s->master_key, s->master_key_length,&d1, &p[0]);
    cert_verify_mac(s->master_key, s->master_key_length,&d2, 
		    &p[MD5_DIGEST_LENGTH]);
  }

#ifdef DEBUG_DIGEST
  print_buffer(p, EVP_MAX_MD_SIZE*2, "my digest: ");
  print_buffer(s->c_v, s->s_v_l, "signature: ");
#endif

}

/** this function is used to calculate the final mac for SSLv3 handshake
 **/
static void cert_verify_mac(unsigned char *master_key, int master_key_length, 
			    EVP_MD_CTX *in_ctx, unsigned char *p) {

  static unsigned char ssl3_pad_1[48]={
    0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,
    0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,
    0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,
    0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,
    0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,
    0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36 };
  
  static unsigned char ssl3_pad_2[48]={
    0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,
    0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,
    0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,
    0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,
    0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,
    0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c,0x5c };
  unsigned int ret;
  int npad,n;
  unsigned int i;
  unsigned char md_buf[EVP_MAX_MD_SIZE];
  EVP_MD_CTX ctx;
  
  EVP_MD_CTX_copy(&ctx,in_ctx);
  
  n=EVP_MD_CTX_size(&ctx);
  npad=(48/n)*n;

  EVP_DigestUpdate(&ctx,master_key,master_key_length);
  EVP_DigestUpdate(&ctx,ssl3_pad_1,npad);
  EVP_DigestFinal(&ctx,md_buf,&i);
  
  EVP_DigestInit(&ctx,EVP_MD_CTX_md(&ctx));

  EVP_DigestUpdate(&ctx,master_key,master_key_length);
  EVP_DigestUpdate(&ctx,ssl3_pad_2,npad);
  EVP_DigestUpdate(&ctx,md_buf,i);
  EVP_DigestFinal(&ctx,p,&ret);
  
  memset(&ctx,0,sizeof(EVP_MD_CTX));

}

/** this function retrieve a certificate out of the buffer (handshake bytes)
 ** the code for extracting the certificate is taken from openssl library
 ** after the certificate is extracted, i verify the signature of the CA
 **/
#define n2l3(c,l)       ((l =(((unsigned long)(c[0]))<<16)| \
                             (((unsigned long)(c[1]))<< 8)| \
                             (((unsigned long)(c[2]))    )),c+=3)

static int retrieve_cert(int v, unsigned char *p, unsigned int plen, 
			   EVP_PKEY **pkey, int *type, X509 **cert) {

  STACK_OF(X509) *sk=NULL;
  X509 *x=NULL;
  unsigned long l,nc,slen;
  unsigned char *q;

  if((sk=sk_X509_new_null()) == NULL) {
    log(TGSLOG_CRIT, "3.1(2): error while getting client's certificate\n");
    return INVALID_HANDSHAKE;
  }

  n2l3(p, slen);

  if(slen+3 != plen) {
    log(TGSLOG_CRIT, "3.1(2): decoding error while getting certificate\n");
    return INVALID_HANDSHAKE;
  }

  for (nc=0; nc<slen; ) {
    n2l3(p,l);
    if ((l+nc+3) > slen) {
      log(TGSLOG_CRIT, "3.1(2): error1\n");
      return INVALID_HANDSHAKE;
    }
    
    q=p;
    x=d2i_X509(NULL,&p,l);
    if (x == NULL) {
      log(TGSLOG_CRIT, "3.1(2): error2\n");
      return INVALID_HANDSHAKE;
    }
    if (p != (q+l)) {
      log(TGSLOG_CRIT, "3.1(2): error3\n");
      return INVALID_HANDSHAKE;
    }
    if (!sk_X509_push(sk,x)) {
      log(TGSLOG_CRIT, "3.1(2): error4\n");
      return INVALID_HANDSHAKE;
    }
    x=NULL;
    nc+=l+3;
  }

  if (sk_X509_num(sk) <= 0 && v == SSL3_VERSION) {
    /* TLS does not mind 0 certs returned */
    log(TGSLOG_CRIT, "3.1(2): certificate list is 0\n");
    return INVALID_HANDSHAKE;
  }

  (*cert) = sk_X509_shift(sk);
  (*pkey) = X509_get_pubkey((*cert));
  (*type) = X509_certificate_type((*cert), (*pkey));


  /** Now need to verify the certificate chain **/

  {
    X509_STORE *certstore = NULL;
    X509_STORE_CTX *certstorectx = NULL;
    X509_LOOKUP *lookup = NULL;
    
    char *CAFile = CA_FILE;
    char *CAPath = CA_PATH;
    int i;
    
    ERR_load_crypto_strings();
    signal(SIGPIPE,SIG_IGN);
    CRYPTO_malloc_init(); 
    SSLeay_add_all_algorithms();
    
    if((certstore = X509_STORE_new()) == NULL) {
      fprintf(stderr, "3.1(2): couldn't create X509_STORE\n");
      return INTERNAL_ERROR;
    }
    
    X509_STORE_set_verify_cb_func(certstore,cb_local);
        
    lookup=X509_STORE_add_lookup(certstore,X509_LOOKUP_file());
    if (lookup == NULL) return INTERNAL_ERROR;
    if (CAFile) {
      i=X509_LOOKUP_load_file(lookup,CAFile,X509_FILETYPE_PEM);
      if(!i) {
	fprintf(stderr, "3.1(2): Error loading file %s\n", CAFile);
	return INTERNAL_ERROR;
      }
    } else X509_LOOKUP_load_file(lookup,NULL,X509_FILETYPE_DEFAULT);
    
    lookup=X509_STORE_add_lookup(certstore,X509_LOOKUP_hash_dir());
    if (lookup == NULL) return INTERNAL_ERROR;
    if (CAPath) {
      i=X509_LOOKUP_add_dir(lookup,CAPath,X509_FILETYPE_PEM);
      if(!i) {
	fprintf(stderr, "3.1(2): Error loading directory %s\n", CAPath);
	return INTERNAL_ERROR;
      }
    } else X509_LOOKUP_add_dir(lookup,NULL,X509_FILETYPE_DEFAULT);

    if((certstorectx = X509_STORE_CTX_new()) == NULL) {
      fprintf(stderr, "3.1(2): could't get new context\n");
      return INTERNAL_ERROR;
    }

    X509_STORE_CTX_init(certstorectx, certstore, (*cert), NULL);
    i = X509_verify_cert(certstorectx);
    X509_STORE_CTX_cleanup(certstorectx);
    if (!i) {
      fprintf(stderr, "retrieve_cert: cert verify failed\n");
      return INVALID_HANDSHAKE;
    }
  }

  return NO_ERROR;
}

/** this function is used during debugging. if verification of certificate
 ** fails, this function will be called
 **/
static int cb_local(int ok, X509_STORE_CTX *ctx)
{
  char buf[256];
  static int v_verbose=0;
 
  if (!ok) {
    X509_NAME_oneline(
		      X509_get_subject_name(ctx->current_cert),buf,256);
    fprintf(stderr, "3.1(2): %s\n",buf);
    fprintf(stderr, "3.1(2): error %d at %d depth lookup:%s\n",ctx->error,
	   ctx->error_depth,
	   X509_verify_cert_error_string(ctx->error));
    if (ctx->error == X509_V_ERR_CERT_HAS_EXPIRED) {
      fprintf(stderr, "expired cert\n");
      ok=1;
    }
    /* since we are just checking the certificates, it is
     * ok if they are self signed. But we should still warn
     * the user.
     */
    if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) {
      fprintf(stderr, "self signed\n");
      ok=1;
    }
    /* Continue after extension errors too */
    if (ctx->error == X509_V_ERR_INVALID_CA) {
      fprintf(stderr, "invalid ca\n");
      ok=1;
    }
    if (ctx->error == X509_V_ERR_PATH_LENGTH_EXCEEDED) {
      fprintf(stderr, "path length exceeded\n");
      ok=1;
    }
    if (ctx->error == X509_V_ERR_INVALID_PURPOSE) {
      fprintf(stderr, "invalid purpose\n");
      ok=1;
    }
    if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) {
      fprintf(stderr, "depth 0 self signed cert\n");
      ok=1;
    }
  }
  if (!v_verbose) ERR_clear_error();
  return(ok);
}

/** this code verifies client's signature within the handshake
 ** code is taken from openssl library (unmodified)
 **/
#define n2s(c,s)        ((s=(((unsigned int)(c[0]))<< 8)| \
                            (((unsigned int)(c[1])))),c+=2)


static int verify_client_signature(unsigned char *p, unsigned int plen, 
				   unsigned char *verify_md, EVP_PKEY *pkey){
  int i, j, ilen, pkey_size;

  n2s(p,ilen);
  plen-=2;

#ifdef DEBUG
  fprintf(stderr, "3.3: verifying the validity of client's signature\n");
#endif

  if (ilen > plen) {
    log(TGSLOG_CRIT, "3.3: error 1\n");
    return INVALID_HANDSHAKE;
  }
  
  pkey_size=EVP_PKEY_size(pkey);
  if ((ilen > pkey_size) || (plen > pkey_size) || (plen <= 0)) {
    log(TGSLOG_CRIT, "3.3: error2\n");
    return INVALID_HANDSHAKE;
  }
  
#ifndef NO_RSA 
  if (pkey->type == EVP_PKEY_RSA) {
    i=RSA_verify(NID_md5_sha1, verify_md,
		 MD5_DIGEST_LENGTH+SHA_DIGEST_LENGTH, p, ilen, 
		 pkey->pkey.rsa);
    if (i < 0) {
      log(TGSLOG_CRIT, "3.3: SSL_R_BAD_RSA_DECRYPT\n");
      return INVALID_HANDSHAKE;
    }
    if (i == 0) {
      log(TGSLOG_CRIT, "3.3: SSL_R_BAD_RSA_SIGNATURE\n");
      return INVALID_HANDSHAKE;
    }
  }
  else
#endif
#ifndef NO_DSA
    if (pkey->type == EVP_PKEY_DSA) {
      j=DSA_verify(pkey->save_type,&(verify_md[MD5_DIGEST_LENGTH]),
		   SHA_DIGEST_LENGTH,p,i,pkey->pkey.dsa);
      if (j <= 0) {
	log(TGSLOG_CRIT, "3.3: DSA_BAD_SIGNATURE\n");
	return INVALID_HANDSHAKE;
      }
    }
    else
#endif
      {
	log(TGSLOG_CRIT, "3.3: invalid encryption type\n");
	return INVALID_HANDSHAKE;
      }
  
  return NO_ERROR;
}


/** Looking up the timestamp in client hello message. 
 ** Looking up the current time (in secs)
 ** Verifying that the difference is within VALID_TIME period
 **/
#define n2l(c,l)        (l =((unsigned long)(*((c)++)))<<24, \
                         l|=((unsigned long)(*((c)++)))<<16, \
                         l|=((unsigned long)(*((c)++)))<< 8, \
                         l|=((unsigned long)(*((c)++))))
static int verify_client_timestamp(unsigned char *p) { 
  unsigned char cTime[SSL3_RANDOM_SIZE];
  unsigned char *pTime;
  unsigned long rTime;
  unsigned long currenttime;

  memcpy(cTime, p, SSL3_RANDOM_SIZE);
  pTime = cTime;
  n2l(pTime, rTime);

  currenttime = time(NULL);

  if(abs(currenttime - rTime)  > VALID_TIME * 60) {
    log(TGSLOG_CRIT, "expired: client=%s current=%s\n", 
	get_time(rTime), get_time(currenttime));
    return OLD_HANDSHAKE;
  }

  return NO_ERROR;
}

static int retrieve_identity(tgs_context *context, char **id, 
			     char **realm, X509 *cert)  {


  X509_NAME *client;
  char      *tmp;

#ifdef DEBUG
  fprintf(stderr, "3.5: retrieve client's identity from the certificate\n");
#endif
  
  client = X509_get_subject_name(cert);
  tmp = (char *)calloc(MAX_BUF_SIZE, sizeof(char));
  (*id) = (char *)calloc(MAX_ID_LEN, sizeof(char));
  (*realm) = (char *)calloc(MAX_REALM_LEN, sizeof(char));

  client = X509_get_subject_name(cert);
  X509_NAME_get_text_by_NID(client, NID_pkcs9_emailAddress, tmp, MAX_BUF_SIZE);
  sscanf(tmp, "%[^@]@%s", (*id), (*realm));
  (*id)[strlen(*id)] = '\0';
  (*realm)[strlen(*id)] = '\0';

  free(tmp);

#ifdef DEBUG
  fprintf(stderr, "3.5: id=%s realm=%s\n", (*id), (*realm));
#endif

  return NO_ERROR;
}

static void process_exit(int sig, const struct sigaction *act,
			   struct sigaction *oact) {

  fprintf(stderr, "process received a wake up call...\n");
  exit(1);
}

static unsigned char *read_message(unsigned char **p, unsigned int *len) {
  unsigned char *buf;
  int hlen = sizeof(unsigned long);

  memcpy((char *)len, (*p), hlen);
  (*len) = ntohl((*len));


#ifdef DEBUG
  fprintf(stderr, "read_message: %d\n", *len);
#endif

  /** check that length is within reasonable range **/
  if((*len) <= 0 || (*len) > MAX_MEMORY_ALLOC) {
    (*p) += hlen;
    return NULL;
  }

  (*p) += hlen;
  buf = (unsigned char *)get_memory((*len), sizeof(unsigned char));
  memcpy(buf, (*p), (*len));
  (*p) += (*len);

  return buf;
}

/*
  static const struct {
    char *name;
    int   nid;
  } cert_struct[] = {
    { "C",     NID_countryName            },
    { "ST",    NID_stateOrProvinceName    }, 
    { "SP",    NID_stateOrProvinceName    }, 
    { "L",     NID_localityName           },
    { "O",     NID_organizationName       },
    { "OU",    NID_organizationalUnitName },
    { "CN",    NID_commonName             },
    { "T",     NID_title                  },
    { "I",     NID_initials               },
    { "G",     NID_givenName              },
    { "S",     NID_surname                },
    { "D",     NID_description            },
    { "UID",   NID_uniqueIdentifier       },
    { "Email", NID_pkcs9_emailAddress     },
    { NULL,    0                          }
  };
*/
