#include "util.h"

void free_memory(void *memory) {

  free(memory);

}

void *get_memory(size_t nelem, size_t elsize) {

  void *m;

  if ((m = (void *)calloc(nelem, elsize)) == NULL) {
    fprintf(stderr, "allocating: %d * %d\n", nelem, elsize);
    perror("calloc failed");
    exit(1);
  }

  return m;

}

int open_log() {

  close(2);

  f = fopen("tgs.log", "a");
  if(!f) return (-1);
  fprintf(f, "[%d]: log is opened\n", getpid());
  fflush(f);

  return 1;
}

void close_log() {   fclose(f); }

char *get_time(time_t t) {
  char *time_string;

  time_string = ctime(&t);
  time_string[strlen(time_string) - 1] = '\0';
  return (time_string);
}

int log(int n, char *fmt, ...) {

     va_list args;
     char *p, *sval;
     int ival;
     double dval;
     int i;

     if(!f) return (-1);

#ifdef TIME
     fprintf(f, "[%d] [%s] -- ", getpid(), get_time());
#endif

     fprintf(f, "[%d]  ", getpid());

     for(i=0; i<n; i++) fprintf(f, " ");
     fprintf(f, "-> ");

     va_start(args, fmt);

     for(p = fmt; *p; p++) {
       if(*p != '%') {
         fprintf(f, "%c", (*p));
         continue;
       }

       switch(*++p) {
       case 'c':
         fprintf(f, "%c", (*p));
         break;
       case 'd':
         ival = va_arg(args, int);
         fprintf(f, "%d", ival);
         break;
       case 'f':
         dval = va_arg(args, double);
         fprintf(f, "%f", dval);
         break;
       case 'x':
	 ival = va_arg(args, int);
	 fprintf(f, "0x%02x", ival);
	 break;
       case 's':
         for(sval = va_arg(args, char *); *sval; sval++)
           fprintf(f, "%c", (*sval));
         break;
       case 'p':
	 ival = va_arg(args, int);
	 fprintf(f, "%p", ival);
	 break;
       default:
         fprintf(f, "%c", (*p));
         break;
       }
     }

     va_end(args);

     fflush(f);

     return 1;
}

size_t recvblock(int sd, unsigned char *buf, size_t total) {

  int len;
  unsigned int recvd = 0;

  while(recvd < total) {
    len = recv(sd, (char *)buf+recvd, total-recvd, 0);
    if(len <= 0) return (-1);
    recvd+=len;
  }
  return (recvd);
}

size_t sendblock(int sd, unsigned char *buf, size_t total) {

  int len;
  unsigned int sent = 0;

  while(sent < total) {
    len = send(sd, (char *)buf+sent, total-sent, 0);
    if(len <= 0) return (-1);
    sent+=len;
  }
  return (sent);
}

void print_buffer(unsigned char *buffer, size_t len, unsigned char *comment) {

  int i;

  fprintf(stderr, "%s ", comment);
  for(i = 0; i < len; i++) {
    fprintf(stderr, "%02X ", buffer[i]);
    if((i+1) % 22 == 0 && i != 0) fprintf(stderr, "\n");
  }
  fprintf(stderr, "\n");

}
